"""
CharaChorder Serial API Wrapper
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

A Python wrapper for CharaChorder's Serial API.

:copyright: (c) 2023-present GetPsyched
:license: MIT, see LICENSE for more details.

"""

__title__ = "charachorder"
__author__ = "GetPsyched"
__license__ = "MIT"
__copyright__ = "Copyright 2023-present GetPsyched"
__version__ = "0.3.1"

from .device import *
from .errors import *
from .types import *
